// <Copyright Holder>. Copyright (C) <Copyright Year(s)>. <License>.
#ifndef HEADER_SRC_APPINFO_H_INCLUDED
#define HEADER_SRC_APPINFO_H_INCLUDED

#include <boost/preprocessor/stringize.hpp>

/*
 * AUTO-GENERATION WARNING:
 *     This file has been automatically generated from "appinfo.h.in.cmake".
 *     DO NOT edit this file, as any changes will be automatically discarded.
 */
 
#define APPLICATION_NAME               "Haplotyping with Recombinations, Errors, and Missing Genotypes"
#define APPLICATION_CODENAME           "reHCstar"
#define APPLICATION_COPYRIGHT_YEARS    "2010, 2011"
#define APPLICATION_VERSION_STRING     "v2.1.1-0-g810fc59"
#define APPLICATION_VENDOR_ID          "org.tecnoparco"
#define APPLICATION_VENDOR_NAME        "Parco Tecnologico Padano"
#define APPLICATION_VENDOR_URL         "www.tecnoparco.org"
#define APPLICATION_ID                 "org.tecnoparco"
#define APPLICATION_SOURCE_VERSION     "v2.1.1-0-g810fc59"

#ifndef APPLICATION_NAME
#   error "Please specify the application name in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_CODENAME
#   error "Please specify the application codename in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_COPYRIGHT_YEARS
#   error "Please specify the application copyright years (years during which this application was created, edited, or modified) in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_VERSION_STRING
#   error "Please specify the application version in the top-level CMakeLists.txt"
#endif

#ifndef APPLICATION_VENDOR_ID
#   error "Please specify the application vendor id in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_VENDOR_NAME
#   error "Please specify the application vendor name in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_VENDOR_URL
#   error "Please specify the application vendor URL in the top-level CMakeLists.txt file."
#endif

#ifndef APPLICATION_ID
#   define APPLICATION_ID APPLICATION_VENDOR_ID "." APPLICATION_CODENAME
#endif

#ifndef APPLICATION_SOURCE_VERSION
#   error "Please specify the application source version in the top-level CmakeLists.txt file or include the ExtractSourceVersionFromGit module."
#endif

#endif
